<?php
/**
 * Projekt: WS-Reporter
 *          Funktionen zur Übersetzung des WS-Reporters
 *
 * Autor:    Wilfried Schollenberger
 *           WS Unternehmensberatung und Controlling-Systeme GmbH
 * Kontakt:  wisch@ws-unternehmensberatung.de
 * Lizenz:   GPL 3.0
 *
 * Stand:    5.8.2022
 *
 *
 */

if (session_status() == PHP_SESSION_NONE) session_start();

/*
 Der IntlDateFormatter hat eine etwas eigenwillige Auslegung bei den formatangaben SHORT, MEDIUM, LONG. 
 Deshalb werden hier die Datums-Formate in einem Array gespeichert.
 */
require_once './includes/funktionen.inc';
require_once './phpsec/createDBH.inc';

class WSRFormats{
 static $wsrDateFormat = array (
    'de' => 'd.m.Y',
    'fr' => 'd/m/Y',
    'en' => 'm/d/Y',
    'en-AUS' => 'd/m/Y' 
  );
 static $wsrTimeFormatS = array (
     'de' => 'H:i:s',
     'fr' => 'H:i:s',
     'en' => 'h.i.s A',
     'en-AUS' => 'h.i.s a'
 );
 static $wsrTimeFormat = array (
     'de' => 'H:i',
     'fr' => 'H:i',
     'en' => 'h.i A',
     'en-AUS' => 'h.i a'
 );
 static $wsrTimeFormatH = array (
     'de' => 'H',
     'fr' => 'H',
     'en' => 'h A',
     'en-AUS' => 'h a'
 );
 static $wsrDateTimeFormatS = array (
     'de' => 'd.m.Y H:i:s',
     'fr' => 'd/m/Y H:i:s',
     'en' => 'm/d/Y h.i.s A',
     'en-AUS' => 'd/m/Y h.i.s a' 
 );

 static $wsrDateTimeFormat = array (
     'de' => 'd.m.Y H:i',
     'fr' => 'd/m/Y H:i',
     'en' => 'm/d/Y h.i A',
     'en-AUS' => 'd/m/Y h.i a'
 );
 static $wsrDateTimeFormatH = array (
     'de' => 'd.m.Y H',
     'fr' => 'd/m/Y H',
     'en' => 'm/d/Y h A',
     'en-AUS' => 'd/m/Y h a'
 );
 
 static $wsrNumberSeparator = array (
     'de' => array(',','.'),
     'fr' => array(',',' '),
     'en' => array('.',',')
 
 );
 
 
 /**
  * Interne Funktion zum Formatieren der Tme- und Datetime-Werte
  * @param DateTime $date
  * @param array $fArray
  * @return string formattierter Wert
  */
  static function formatDT($date,$fArray) {
      $lang = null;
      if (array_key_exists('Sprache', $_SESSION)) $lang = $_SESSION['Sprache'];
      if ($lang == null) $lang = 'de';
      $format = $fArray['de'];
      if (array_key_exists($lang, $fArray)) $format = $fArray[$lang];
      else
          if (array_key_exists(substr($lang,0,2), $fArray)) $format = $fArray[substr($lang,0,2)];
      return $date->format($format);    
      
  }

  
  /**
   * Formatiert einen Datums-Wert nach Sprach-Einstellung
   * @param DateTime $date
   * @return string
   */
  static function formatDate($date) {
      if ($date instanceof DateTime)
        return self::formatDT($date,self::$wsrDateFormat);
      return '';
  }
  
  /**
   * Formatiert einen DateTime-Wert
   * @param DateTime $dateTime
   * @return string
   */
  static function formatDateTime($dateTime) {
      if ($dateTime instanceof DateTime)
          return self::formatDT($dateTime,self::$wsrDateTimeFormat);
     return '';
  }

  /**
   * Formatiert einen DateTime-Wert mit Sekunden
   * @param DateTime $dateTime
   * @return string
   */
  static function formatDateTimeS($dateTime) {
      if ($dateTime instanceof DateTime)
          return self::formatDT($dateTime,self::$wsrDateTimeFormatS);
      return '';
  }
  
  /**
   * Formatiert eine Uhrzeit
   * @param DateTime $time
   * @return string
   */
  static function formatTime($time) {
      if ($time instanceof DateTime)
          return self::formatDT($time,self::$wsrTimeFormat);
      return '';
  }

  /**
   * Formatiert eine Uhrzeit mit Sekunden
   * @param DateTime $time
   * @return string
   */
  static function formatTimeS($time) {
      if ($time instanceof DateTime)
          return self::formatDT($time,self::$wsrTimeFormatS);
          return '';
  }
  
  /**
   * Formatiert eine Dezimalzahl
   * @param double $wert
   * @param int $decs
   * @return string
   */
  static function formatDecimal($wert,$decs){
      $lang = null;
      if (array_key_exists('Sprache', $_SESSION)) $lang = $_SESSION['Sprache'];
      if ($lang == null) $lang = 'de';
      $seps = self::$wsrNumberSeparator['de'];
      if (array_key_exists($lang, self::$wsrNumberSeparator))  $seps = self::$wsrNumberSeparator[$lang];
      else
          if (array_key_exists(substr($lang,0,2), self::$wsrNumberSeparator))  $seps = self::$wsrNumberSeparator[substr($lang,0,2)];
     return number_format($wert,$decs,$seps[0],$seps[1]);
  }
 
  
  /**
   * Interne Funktion erzeugt DateTime-Objekt aus Eingabe
   * @param string value
   * @param array $fArray
   * @return DateTime 
   */
  static function createDT($value,$fArray) {
      $lang = null;
      if (array_key_exists('Sprache', $_SESSION)) $lang = $_SESSION['Sprache'];
      if ($lang == null) $lang = 'de';
      $format = $fArray['de'];
      if (array_key_exists($lang, $fArray)) $format = $fArray[$lang];
      else
          if (array_key_exists(substr($lang,0,2), $fArray)) $format = $fArray[substr($lang,0,2)];
      $dt =  DateTime::createFromFormat($format,$value);          
      if ($dt === false) {
          error_log( 'FEHLER: ungültiges Datum '.$value.' für Sprache '.$lang.' mit Format '.$format);          
      }else{
          error_log($value.' ergibt mit '.$format.' '.$dt->format('Y-m-d H:i:s'));
      }
      return $dt;
  }

  /**
   * Liest Date-Wert in aktueller Sprache und formatiert für Datenbank (Y-m-d)
   * @param string $value Datum
   * @return string or false
   */
  static function stdDate($value){
     $dt = self::createDT($value, self::$wsrDateFormat);
     if ($dt === false) return false;
     return $dt->format('Y-m-d');
  }
  
  /**
   * Liest Datetime-Wert in aktueller Sprache und formatiert für Datenbank (Y-m-d H:i:s)
   * @param string $value Datum
   * @return string or false
   */
  static function stdDateTime($value){
      $dt = self::createDT($value, self::$wsrDateTimeFormatS);
      error_log('??DT '.print_r($dt,true));
      if ($dt === false) 
          $dt = self::createDT($value, self::$wsrDateTimeFormat);
      if ($dt === false)
          $dt = self::createDT($value, self::$wsrDateTimeFormatH);
      if ($dt === false) return false;
      return $dt->format('Y-m-d H:i:s');
  }
  
  /**
   * Liest Time-Wert in aktueller Sprache und formatiert für Datenbank (H:i:s)
   * @param string $value Datum
   * @return string or false
   */
  static function stdTime($value){
      $dt = self::createDT($value, self::$wsrTimeFormatS);
      if ($dt === false) $dt = self::createDT($value, self::$wsrTimeFormat);
      if ($dt === false) $dt = self::createDT($value, self::$wsrTimeFormatH);
      if ($dt === false) return false;
      return $dt->format('H:i:s');
  }
 
  /**
   * Liest eine Zahl in aktueller Sprache und formatiert für Datenbank mit Dezimalpunkt
   * @param string $value
   * @return double
   */
  static function stdNumber($value){
      $lang = $_SESSION['Sprache'];
      if ($lang == null) $lang = 'de';
      $seps = self::$wsrNumberSeparator['de'];
      if (array_key_exists($lang, self::$wsrNumberSeparator))  $seps = self::$wsrNumberSeparator[$lang];
      else
          if (array_key_exists(substr($lang,0,2), self::$wsrNumberSeparator))  $seps = self::$wsrNumberSeparator[substr($lang,0,2)];
      $value = str_replace($seps[1],'', $value);
      if ($seps[0] != '.' ) $value = str_replace($seps[0],'.', $value);
      return doubleval($value);
  }  
}

class WSRTranslate{
  /**
   * Übersetzt statische Texte des WS-Reporters
   * @param string $text
   * @return string Übersetzung
   */
  static function translate($text){
      global $dbh;
      static $countStatement = null;
      static $readStatement = null;
      static $insertStatement = null;
      $text = html_entity_decode($text);
      $text = trim(str_replace("\r", '',$text));
      
      $sprache = null;
      if (array_key_exists('Sprache', $_SESSION)) $sprache = substr($_SESSION['Sprache'],0,2);
      if (! isValid($sprache)) $sprache = 'de';
//nur bei Bedarf      error_log('Übersetze '.$text);
      if ($sprache == 'de' && $_SERVER["HTTP_HOST"] != "localhost:8898") return $text;
      if ($readStatement == null){
          $statement = 'select '.$sprache.' as trans from wsrTranslations where de = :text';
          $readStatement = $dbh->prepare($statement);
          if ($readStatement === false){
              echo "self WS-Reporter FEHLER:\n".$dbh->errorInfo()[2]."\n".$statement;
              die();
          }
      }
      if (false === $readStatement->execute (array('text'=> $text) )){
          echo "self WS-Reporter READ-FEHLER:\n".$readStatement->errorInfo()[2]."\n".$statement;
          echo "\nText=".$text;
          die();
      }
      $z = $readStatement->fetchObject();
      if ($z == false ) {
          // In Entwicklung Texte sammeln
          if ($_SERVER["HTTP_HOST"] == "localhost:8898"){
              error_log('Sammle '.$text);
              if ($insertStatement == null){
                  $statement = 'insert into wsrTranslations set de = :text';
                  $insertStatement = $dbh->prepare($statement);
                  if ($insertStatement === false){
                      echo "self WS-Reporter FEHLER:\n".$dbh->errorInfo()[2]."\n".$statement;
                      die();
                  }
              }
              if (false === $insertStatement->execute (array('text' => $text))){
                  echo "self WS-Reporter INSERT-FEHLER:\n".$insertStatement->errorInfo()[2]."\n".$statement;
                  echo "\nText=".$text;
                  die();
              }              
          }
          return $text;
      }else{
          if ($countStatement == null){
              $statement = 'update wsrTranslations set zaehler = zaehler+1 where de = :text';
              $countStatement = $dbh->prepare($statement);
              if ($countStatement === false){
                  echo "count WS-Reporter FEHLER:\n".$dbh->errorInfo()[2]."\n".$statement;
                  die();
              }
          }
          if (false === $countStatement->execute (array('text'=> $text) )){
              echo "count WS-Reporter READ-FEHLER:\n".$countStatement->errorInfo()[2]."\n".$statement;
              echo "\nText=".$text;
              die();
          }
      }
      if (! isValid($z->trans)) {
          return $text;
      }
      return $z->trans;
  }
 
  /**
   * Übersetzten und spezielle Zeichen, ", <, >, & maskieren.
   * @param string $text
   * @return string übersetzter und maskierter Text
   */
  static function translateHTML($text){
      $ttext = self::translate($text);
      return htmlspecialchars($ttext);
  }
  
  /**
   * Übersetzten und ggf. Übersetzer-Icon hinzufügen 
   * @param string $text
   * @return string übersetzter Text, ggf.  mit Icon
   */
  static function translateH($text){
      $ttext = self::translate($text);
      if (array_key_exists('TIcons', $_SESSION)){
          $utext = urlencode($text);
          if ($text == $ttext) $icon = "<span class=\"bButton\"";
              else $icon = "<span class=\"gButton\"";
           $icon .= " title='Übersetzung bearbeiten'"
              ."onclick ='window.open(\"wsrTranslateText.php?text=".$utext.'","dispWindow","width=600,height=600")\''
                  ." >T</span>";
                  return $ttext.$icon;
      }
      return $ttext;
  }

  /**
   * Übersetzten und spezielle Zeichen, ", <, >, & maskieren.
   * @param string $text
   * @return string übersetzter und maskierter Text ggf. mit Icon zur Übersetzung
   */
  static function translateHTMLH($text){
      $ttext = self::translate($text);
      if (array_key_exists('TIcons', $_SESSION)){
          $utext = urlencode($text);
          if ($text == $ttext) $icon = "<span class=\"bButton\"";
          else $icon = "<span class=\"gButton\"";
          $icon .= " title='Übersetzung bearbeiten'"
              ."onclick ='window.open(\"wsrTranslateText.php?text=".$utext.'","dispWindow","width=600,height=600")\''
                  ." >T</span>";
                  return htmlspecialchars($ttext).$icon;
      }
      return htmlspecialchars($ttext);
  }
  
}