/**
 * Teile von HTML-Tabellen auf- und zuklappen
 *
 * Autor:    Wilfried Schollenberger
 *           WS Unternehmensberatung und Controlling-Systeme GmbH
 * Kontakt:  wisch@ws-unternehmensberatung.de
 * Lizenz:   LGPL 3.0
 *
 * Stand:    6.6.2022
 *
 * Status: Produktive Version für WS-Reporter.
 *         Kompatibel mit TableHeader. 
 *
 * Das Script definiert eine Klasse TableFolder, die bei 
 * TH-Tags mit rowSpan > 1 alle Zeilen bis auf die erste (im Header-Mode)
 * oder die letzte (im Summen-Mode) aus- und einblendet.
 * Dabei blebt entweder die erste Zeile als Titel-Zeile
 * oder die letzte Zeile als Summenzeile stehen.
 *
 * Dieses Verhalten kann für die gesamte Tabelle oder für einzelne
 * TH-Tags über die Pseudo-Klassen WSFSummen und WSFHeader 
 * gesteuert werden. Auch im Konstruktor kann der Typ Header oder Summen
 * angegeben werden.
 * 
 * Weitere Optonen (bool):
 *  * start: TableFolder.Ausgeklappt, TableFolder.Eingeklappt, TableFolder.Stufe1 oder TableFolder.Stufe2
 *  * globalToggler (bool): Im THEAD-Abschnitt werden Schalter für die
 *    jeweilige Spalte insgesamt angelegt.
 *    Diese Funktion muss explizit über die Klassen WSFSummen oder WSFHeader 
 *    für das jeweilge TH-Tag aktiviert werden.
 *  * iconsInCSS (bool): Wird diese Option gesetzt, dann müssen die auszugebenden Zeichen
 *    in den CSS-Regeln  WSFIconOpen:after (öffnen, aufklappen)
 *    und WSFIconClose:after (schließen, einklappen) mit Content angegeben werden
 *    Wenn diese Option nicht gesetzt ist, werden die 
 *    document.styleSheets[j].cssRules durchsucht.  
 *
 * Für die Schalter werden der CSS-Style WSFIcon, WSFIconOpen (öffnen, aufklappen)
 * und WSFIconClose (schließen, einklappen) verwendet.
 *
 * Um das Ganze im WS-Reporter komfortabel zu machen, prüft der Tabellen-Header, 
 * ob die Styles WSFIconOpen:after und WSFIconClose:after definiert sind.
 * Wenn ja, werden keine Pfeile gesetzt, damit nur die content-Attrubte
 * aus diesen Definitionen wirksam sind. 
 *
 * Wenn die Methode restore() aufgerufen wird, speichert der TableFolder den
 * Zustand der Knoten in windows.sessionStorage. Dabei wird der Text-Inhalt des Knotens
 * verwendet. Beim nächsten Aufruf wird dieser Zustand bei Knoten mit gleichem Inhalt
 * wiederhergestellt.
 * Da es aber auch Knoten ohne eindeutigen Inhalt gibt, muss die komplette 
 * Hierarchie berücksichtigt werden.
 * Beispiel:
 *   var tf = new TableFolder("ReportTabelle",TableFolder.Summen,true,true);
 *   tf.restore();
 *    
 * Unklar: Wird der Typ WSFReporter für Sonderbehandlungen im WS-Reporter
 *         benötigt?
 *         Aktuell nicht.
 *
 * Bugs, Probleme, Hinweise:
 * 
 * 1.) Wenn TableHeader und TableFolder für eine Tabelle verwendet werden, 
 *     muss der TableHeader vor dem TableFolder instanziiert werden.
 *     Aktuell funktioniert das Auf- und Zu-Klappen nicht aus den kopierten Zellen.
 *
 * 2.) Beim Einklappen erhält die Summen-Zeile die Klassen der ersten Zeile der Gruppe
 *     und wird so zu einer "normalen" Zeile. 
 *     In den kopierten Zeilen-Label des TableHeaders wird das nicht übernommen.
 *     Deshalb bleibt z.B. im WS-Reporter das "Insgesamt" farbig unterlegt.
 *
 * Anmerkung des Architekten:-)
 * Wer hier ein Beispiel für "saubere" objektorientierte Programmierung sucht,
 * wird schwer enttäuscht. Einerseits ist es ein großer Vorteil, dass die vorhandenen
 * DOM-Objekte einfach um Attribute erweitert werden können. 
 * Andererseits haben wir darauf verzichtet, für diese Attribute setter-Methoden zu
 * schreiben, die Seiten-Effekte sauber abfangen.
 * Deshalb müssen bei jeder Änderung etwaige Folgen für WSFnexts-, WSFfollows und THcopy-
 * Objekte jeweils neu berücksichtigt werden.
 * Das ist alles andere als "schön".
 * Der Entwickler bittet um Vergebung:-). 
 *
 */;


/**
 * Klasse mit den Methoden zum Auf- und Zuklappen von HTML-Tabellen
 */
class TableFolder{
	/**
	 * Konstanten
	 */
	static NachKlassen = 0;
	static Summen = 2;
	static Header = 3;
	static WSReporter = 1;
	static NichtsMachen = -1;
	static Knoten = 1;
	static Follower = 0;
	static Daten = -1;
	
	static Ausgeklappt = -1;
	static Eingeklappt = 0;
	static Stufe1 = 1;
	static Stufe2 = 2;

	
	/**
	 Konstruktor
     @param name String ID der HTML-Tabelle
     @param typ int Globaler Typ des Einklapp-Modus
     @param start int Beim Start alles einklappen oder nur erste Spalten(n) aufklappen
     @param globalToggler bool Schalter für ganze Spalten in den Spalten-Köpfen
	 @param storeSession bool Der Status der Elemente wird lokal in den Window-Daten
                              gespeichert und bei erneutem Aufruf wieder hergestellt.
	 @param oconsInCSS bool Hinweis, ob die CSS-Regeln WSFIconOpen::after
                            und WSFIconClose::after gesetzt sind. 
							Wenn der Parameter nicht gesetzt ist, werden die
							document.styleSheets[].cssRules durchsucht. 
	 */
	constructor(name,typ,start,globalToggler,storeSession,iconsInCSS){
		if (document.getElementById(name) == null) return;
		console.log('Initialisiere TableFolder für '+name);
		this.name = name;
		this.start = start;
		this.globalToggler = globalToggler;
		this.iconsInCSS = iconsInCSS;
		this.storeSession = storeSession;
		if (iconsInCSS){
			this.iconOpenInCSS = true;
			this.iconCloseInCSS = true;
		}else if (iconsInCSS == null)
			for (let j=0; j < document.styleSheets.length; j++){
				let liste =  document.styleSheets[j].cssRules;
				for (let i = 0; i < liste.length;i++){
					if (liste[i].selectorText == '.WSFIconOpen::after') {
						this.iconOpenInCSS = true;
						console.log('Hinweis: Verwende .WSFIconOpen::after');	
					} 	
					if (liste[i].selectorText == '.WSFIconClose::after'){
						this.iconCloseInCSS = true;
						console.log('Hinweis: Verwende .WSFIconClose::after');	
					}
				}					
			}
		var tabelle = document.getElementById(name);
		tabelle.WSTableFolder = this; 
		if (tabelle.classList.contains('WSFSummen')) this.typ = TableFolder.Summen;
		else
		if (tabelle.classList.contains('WSFHeader')) this.typ = TableFolder.Header;
		else
		if (tabelle.classList.contains('WSFReporter')) this.typ = TableFolder.WSReporter;
		else
			this.typ = typ;
		this.tabelle = tabelle;
		this.thElemente = [];
		this.init('tbody');
//		this.init('tfoot');		
		if (start >= TableFolder.Eingeklappt){
			var storeSession = this.storeSession;
			this.storeSession = false;
			for (let i = this.thElemente.length -1; i >= 0; i--){
				if (! this.thElemente[i].WSFopen) continue;
				var prev = this.thElemente[i].WSFprev;
				while (prev != null && prev.WSFtyp <= TableFolder.Knoten) prev = prev.WSFprev;
				if (prev == null){
					this.doHideRows(this.thElemente[i]);	
					if (start >= TableFolder.Stufe1){
						this.doShowRows(this.thElemente[i]);	
						if (start >= TableFolder.Stufe2){
							for (let j = 0; j < this.thElemente[i].WSFnexts.length; j++){
								this.doShowRows(this.thElemente[i].WSFnexts[j]);	
							}
						}					
					}					
				}
			}
			this.storeSession = storeSession;
		} 	
		if (globalToggler) this.addGlobalToggler();		
		if (this.storeSession) this.restore();
		if (this.tabelle.WSTableHeader != null) this.tabelle.WSTableHeader.resize();		
		console.log('TableFolder für '+name+' ist fertig');
	}
	
	
	/**
	Initialisieren.
	Hier werden die TH-und TR-Elemente mit zusätzlichen WSFxx-Attributen versehen.
	Weiterhin werden die Schalter in den TH-Elementen eingetragen.
	
	Struktur:
	WSFtyp: Th.Summen (2), TH.Header, TH.Knoten, TH.Follower
	WSFrow: Ursprüngliche Row des Elements
	WSFpos: Position für WSFclosePos beim Schließen des Knotens 
	WSFrows: Ursprüngliche RowSpan
	WSFprev: hierarchisch übergeordnetes oder vorhergehendes Element
	WSFnexts: Array mit hierarchisch untergeordneten Elementen (rowspan < WSFRows und WSHTyp < Th.Summen
	WSFfollows: Array mit nachfolgenden Elementen mit gleicher RowSpan
	
	
	
	TH-Elemente erhalten:
	WSFoldClass: className, der beim Schließen geändert und hiermit zurück gesetzt wird	
	THcopy aus TableHeader
	
	Die Struktur wird aufgebaut, bis das letzte TH-Element gefunden wurde. 
	Wenn das eine RowSpan > 1 hat, werden auch noch die Follower ermittelt, aber keine WSNexts.

     @param tablePart String Tabellen-Abschnitt (thead | tbody)
	 */;
	init (tablePart){
		if (tablePart == null) tablePart = 'tbody';
		var tabelle = document.getElementById(this.name);
		var tbody = tabelle.getElementsByTagName(tablePart)[0];

		if (tbody == null){
			alert("Fehler beim Instanziieren des TableFolders:\nDas TBODY-Tag wurde nicht gefunden.");
			return;
		}
		this.tabelle = tabelle;
		this.thElemente = Array();
		
		var row = tbody.firstElementChild;
		while (row != null){
			// NICHT MEHR ?? row.WSFclosePos = -1;
			// Bei Close mit Sumenzeile wird className in der Summenzeile durch className
			// aus der ersten Zeile ersetzt.
			// Beim Open muss wiedr zurück gesetzt werden.
			row.WSFoldClass = row.className;
			var th = row.firstElementChild;
			if (th.rowSpan > 1){
				th.WSFpos = 0;
				this.initObject(th,null);
				for (let i=0; i < th.WSFrows; i++){
					row.WSFoldClass = row.className;
					row = row.nextElementSibling;
				}
			}
			else row = row.nextElementSibling;
		}
	}

	/**
	Einzelnes Objekt initialisieren
	     @param th HTML-Node TH-Node, die erweitert wird
	     @param prevTH HTML-Node TH-Node, die als Vorgänger eingetragen wird
	 */ 
	initObject(th,prevTH){
		var nextPos = th.WSFpos + 1;
		th.WSFprev = prevTH;
		th.WSFrow = th.parentElement;
		th.WSFrow.WSFclosePos = -1;
		if (th.rowSpan > 1) th.WSFrows = th.rowSpan;
		else th.WSFrows = 1;
		if (th.classList.contains('WSFSummen')) th.WSFtyp = TableFolder.Summen;
		else
		if (th.classList.contains('WSFHeader')) th.WSFtyp = TableFolder.Header;
		else 
		if (prevTH != null && th.rowSpan == prevTH.rowSpan) {
			th.WSFtyp = TableFolder.Follower;
			th.WSFprev = prevTH;
			return;
		}
		else
		if (th.rowSpan > 1){
			if (th.tagName == 'TH' && this.typ > TableFolder.Knoten)
				th.WSFtyp = this.typ;
			else
				th.WSFtyp = TableFolder.Knoten;
		} 
		else {
			let text = 'initOpject 1 rowSpan < 2 '+th.innerText;
			console.log(text);
			return;
		}
		th.WSFfollows = Array();
		th.WSFnexts = Array();
		th.WSFopen = true;
		var rows = th.WSFrows;
		var row = th.WSFrow;
		var tb = th.nextElementSibling;
		// Erste Zeile separat bearbeiten, weil sie das th-Element enthält;
		while (tb != null && tb.rowSpan == th.rowSpan){
			let tbPrev = tb.previosElementSibling; 			
			if (tb.classList.contains('WSFSummen')) {
				tb.WSFpos = nextPos;
				th.WSFnexts.push(tb);
				this.initObject(tb,th);	
				tb = null;													
			}
			else
			if (tb.classList.contains('WSFHeader')){
				tb.WSFpos = nextPos;
				th.WSFnexts.push(tb);
				this.initObject(tb,th);						
				tb = null;
			}
			else{
				tb.WSFpos = nextPos;
				nextPos ++;
				th.WSFfollows.push(tb);
				this.initObject(tb,th);		
				tb = tb.nextElementSibling;
			}		 			
		}
		if (tb == null) return;
		
		// Ersten Knoten bearbeiten
		if (tb.rowSpan < 2){
			tb.WSFpos = nextPos;
			th.WSFnexts.push(tb);
			tb.WSFtyp = TableFolder.Daten;
			row = row.nextElementSibling;
			rows -= 1;
		}else{
			tb.WSFpos = nextPos;
			th.WSFnexts.push(tb);
			this.initObject(tb,th);
			rows -= tb.WSFrows;
			for (let i = 0; i < tb.WSFrows; i++)
				row = row.nextElementSibling;
		}
		
		while (rows > 0){
			tb = row.firstElementChild;
			if (tb.rowSpan < 2){
				tb.WSFpos = nextPos;
				th.WSFnexts.push(tb);
				tb.WSFtyp = TableFolder.Daten;
				row = row.nextElementSibling;
				rows -= 1;
			}else{
				tb.WSFpos = nextPos;
				th.WSFnexts.push(tb);
				this.initObject(tb,th);
				rows -= tb.WSFrows;
				for (let i = 0; i < tb.WSFrows; i++)
					row = row.nextElementSibling;
			}
		}
		if (rows < 0) {
			let text = 'initObject 2 rows = '+rows+' bei '+th.innerText;
			console.log(text);	
		}
		
		// Jetzt noch den Pfeil einfügen
		if (th.WSFtyp > TableFolder.Knoten){
			if (th.firstChild.nodeType == Node.ELEMENT_NODE) th.firstChild.style.display = 'inline';
			var icon =  this.newIconZu(th);
			icon.addEventListener('click',this.hideRows);
			th.insertBefore(icon,th.firstChild);
			// Zusätzliche Attribute im th-Element
			this.thElemente.push(th);
		}
		return;
	} // Ende Init Object




	/**
	 * Schalter in den Spaltenköpfen erzeugen, wenn globalToggler gesetzt ist.
	 */
	addGlobalToggler(){
		var tabelle = document.getElementById(this.name);
		var thead = tabelle.getElementsByTagName('thead')[0];
		var row = thead.firstElementChild;
		var wsfPos = 0;
		var wsfStufe = 0;
		var wsfPosArray = [0];
		var wsfPrevArray = [];
		var icon = null;
		if (row != null){
			this.thHeadList = row.getElementsByTagName('th');
			for (let i = 0; i < this.thHeadList.length; i++){				
				var th = this.thHeadList[i];				
				// Typ bestimmen
				if (th.classList.contains('WSFSummen')) th.WSFtyp = TableFolder.Summen;
				else
				if (th.classList.contains('WSFHeader')) th.WSFtyp = TableFolder.Header;
				else th.WSFtyp = TableFolder.NichtsMachen;

				if (th.WSFtyp > 1){
					// Icon zum Zuklappen in TH-Element einfügen, Zeilenumbruch wg. neuem Node vermeiden
					if (th.firstChild.nodeType == Node.ELEMENT_NODE) th.firstChild.style.display = 'inline';
					if (this.start >= TableFolder.Eingeklappt
						&& wsfStufe >= this.start) {
						icon =  this.newIconAuf(th);
						th.WSFopen = false;
						icon.addEventListener('click',this.showColRows);						
					}else{
						icon =  this.newIconZu(th);						
						th.WSFopen = true;
						icon.addEventListener('click',this.hideColRows);
						wsfStufe ++;
					}
					th.insertBefore(icon,th.firstChild);

					// Zusätzliche Attribute im th-Element
					th.WSFpos = i; 						
					th.WSFTableFolder = this; 	
				}
				else {
					wsfPosArray[wsfPos] = 1;
					th.WSFtyp = TableFolder.NichtsMachen;	
				}
			}
		}
		return;
	}


	/**
	Zeilen einklappen
	@param thElement Element, dessen Zeilen eingeklappt (verborgen) werden
	@param all bool nachfolgende Elemente ebenfalls einklappen (zur Zeit nicht implementiert)
	 */
	doHideRows(thElement,all){
		var text=''; // zum Debuggen;
		var thb = null;		
		if (!thElement.WSFopen) return;
		if (thElement.WSFnexts == null) return;
		if (thElement.WSFnexts.length < 1) return;
		// Untergeordnete Elemente schließen  
		for (let i = 0; i < thElement.WSFnexts.length; i++){
			if (thElement.WSFnexts[i].WSFopen && thElement.WSFnexts[i].rowSpan > 1) this.doHideRows(thElement.WSFnexts[i]);
		}
		if (thElement.WSFtyp <= TableFolder.Knoten) return;
		thElement.WSFopen = false;

		var row = thElement.WSFrow;
		var rows = thElement.WSFrows;
		var firstRow = null;
		var lastRow = null;
		var children = null;
		var count = 0;
		
		// Noch sichtbare Zeilen schließen
		for (let i = 0; i < rows; i++){
			if (! row.hidden){
				if (firstRow == null) {
					firstRow = row;
				}
				
				if (thElement.WSFtyp == TableFolder.Summen && firstRow != row) {
					
					// lastRow-Zeile löschen. Damit bleibt die letzte Zeile stehen.
					// Dabei sollen Knoten in der Summenzeile als Einheit behandelt werden
					// Deshalb werden unten bei Knoten (rowSpan > 1) nachfolgende Zeilen 
					// überlesen und müssen dann hier en block bearbeitet werden
					let fe = lastRow.firstElementChild;
					if (lastRow == firstRow && thElement.WSFnexts.length > 0) fe = thElement.WSFnexts[0]
					if (fe.rowSpan > 1){
						let kRow = lastRow;
						let knotenRows = fe.WSFrows;
						for (let j = 0; j < knotenRows; j++){
							if (! kRow.hidden){
								kRow.hidden = true;
								kRow.WSFclosePos = thElement.WSFpos;
								if (kRow.THcopy != null) kRow.THcopy.hidden = true;
								count ++; 								
							}
							kRow = kRow.nextElementSibling;
						}						
					}else{
						lastRow.hidden = true;
						lastRow.WSFclosePos = thElement.WSFpos;
						if (lastRow.THcopy != null) lastRow.THcopy.hidden = true;
						count ++; 								
					}
				}else				
				if (thElement.WSFtyp == TableFolder.Header && firstRow != row){
					row.hidden = true;
					row.WSFclosePos = thElement.WSFpos;
					if (row.THcopy != null) row.THcopy.hidden = true;
					count ++; 								
				}
				lastRow = row;
			}
			if (firstRow == row
			    && thElement.WSFnexts != null
				&& thElement.WSFnexts.length > 0
				){ 
					if (thElement.WSFnexts[0].rowSpan > 1){
					let fe = thElement.WSFnexts[0];
					let knotenRows = fe.WSFrows;
					let wsfRow = row.firstElementChild.WSFrow;
					let startRow = row;
					for (let j = 0; j < knotenRows; j++){
						if (row == null){
							text = 'FATAL beim Überspringen '+j+' '+thElement.innerText;
							console.log(text);							
						}else{
							if (startRow == wsfRow){
								row = row.nextElementSibling;
								i++;
							}else wsfRow = wsfRow.nextElementSibling;
						}
					}
					i--;
					}else 
							row = row.nextElementSibling;
			}else
			if (row.firstElementChild.rowSpan > 1 && thElement.WSFtyp == TableFolder.Summen){
				let knotenRows = row.firstElementChild.WSFrows;
				let wsfRow = row.firstElementChild.WSFrow;
				let startRow = row;
				for (let j = 0; j < knotenRows; j++){					
					if (row == null){
						text = 'FATAL bei Überspringen '+j+' '+thElement.innerText;
						console.log(text);
					}else{
						if (startRow == wsfRow){
							row = row.nextElementSibling;
							i ++;
						}else 
							wsfRow = wsfRow.nextElementSibling;
					}
				}
				i--;
			}else 	
				row = row.nextElementSibling; 
		}

		var tha = thElement;
		var THcopyCount = 0;

		// rowSpan für das Element, seine Follower 
		// und hierarchisch vorhergehende Elemente mit ihren Follower setzen			
		while (tha != null){
			tha.rowSpan -= count;
			if (tha.THcopy != null) tha.THcopy.rowSpan -= count;	
			for (let i = 0; i < tha.WSFfollows.length; i++){
				thb = tha.WSFfollows[i];
				thb.rowSpan -= count;
				if (thb.THcopy != null) thb.THcopy.rowSpan -= count; 
			}
			tha = tha.WSFprev;			
		}

		if (thElement.WSFtyp == TableFolder.Summen){
			// Bei Typ Summen Elemente in LastRow verschieben	
				
			var thStart = lastRow.firstElementChild;
			var thStartCopy = null;		
			if (lastRow.THcopy != null) thStartCopy = lastRow.THcopy.firstElementChild;
			var THcopyCount = 0;
			do{
				thb = firstRow.firstElementChild;				
				thStart.before(thb);
				if (thb.THcopy != null){
					if (thStartCopy != null) thStartCopy.before(thb.THcopy);
					else lastRow.THcopy.append(thb.THcopy);
					THcopyCount ++;			
				}
			}while (thb != thElement)

			for (let i = 0; i < thElement.WSFfollows.length;i++){
				thb = thElement.WSFfollows[i];				
				thStart.before(thb);
				if (thb.THcopy != null){
					if (thStartCopy != null) thStartCopy.before(thb.THcopy);
					else lastRow.THcopy.append(thb.THcopy);
					THcopyCount ++;			
				}			
			}
		
			if (THcopyCount > 0){
				if (this.tabelle.WSTableHeader != null)
					this.tabelle.WSTableHeader.TFkopiere(THcopyCount,firstRow.THcopy,lastRow.THcopy);
			}
			lastRow.className = firstRow.className; 
		}
		
		let icon = this.newIconAuf(thElement); 
		icon.addEventListener('click',this.showRows);
		thElement.replaceChild(icon,thElement.firstChild);
		if (this.storeSession){
			window.sessionStorage.setItem(this.storeText(thElement),'closed');
		}

/*		
		// Kopie aus TableHeader funktioniert nicht
		if (thElement.THCopy != null){
			let thc = thElement.THCopy;
			let icon = this.newIconAuf(thElement); 
			icon.addEventListener('click',this.showRows);
			thc.replaceChild(icon,thc.firstChild);
		}
// */		
		return;
	}
	


	/**
	Zeilen aufklappen
	@param thElement Element, dessen Zeilen wieder angezeigt werden
	@param all bool nachfolgende Elemente ebenfalls aufklappen (zur Zeit nicht implementiert)
	 */
	doShowRows(thElement,all){
		if (thElement.WSFopen) return;

		// Übergeordnete Elemente öffnen (läuft rekursiv)
		if (thElement.WSFprev != null) this.doShowRows(thElement.WSFprev);

		if (thElement.WSFtyp <= TableFolder.Knoten) return;
		  
		var text = '';
		var thb = null;
		var row = thElement.WSFrow;
		var rows = thElement.WSFrows;
		var firstRow = null;
		var lastRow = null;
//		var children = null;
		var count = 0;

		thElement.WSFopen = true;

		// Zeilen wieder anzeigen
		for (let i = 0; i < rows; i++){
			if (row.hidden && row.WSFclosePos <= thElement.WSFpos) {
				row.WSFclosePos = -1;
				row.hidden = false;	
				if (firstRow == null) firstRow = row;
				if (row.THcopy != null) row.THcopy.hidden = false;
				count ++;				 
			}else if (! row.hidden && row == thElement.parentElement) {
				lastRow = row;
			}
			row = row.nextElementSibling;
		}
				
		var tha = thElement;
		var THcopyCount = 0;

		// rowSpan für Element und hierarchisch vorhergehende Elemente setzen			
		while (tha != null){
			tha.rowSpan += count;
			if (tha.THcopy != null) tha.THcopy.rowSpan += count;	
			for (let i = 0; i < tha.WSFfollows.length; i++){
				thb = tha.WSFfollows[i];
				thb.rowSpan += count;
				if (thb.THcopy != null) thb.THcopy.rowSpan += count; 
			}
			tha = tha.WSFprev;			
		}
		
				
		if (thElement.WSFtyp == TableFolder.Summen){	
			// Bei Typ Summen Elemente aus lastRow nach firstRow verschieben		
			var thStart = firstRow.firstElementChild;
			var thStartCopy = null;		
			if (firstRow.THcopy != null) thStartCopy = firstRow.THcopy.firstElementChild;
			var THcopyCount = 0;
			do{
				thb = lastRow.firstElementChild;	
				thStart.before(thb);
				if (thb.THcopy != null){
					if (thStartCopy != null) thStartCopy.before(thb.THcopy);
					else firstRow.THcopy.append(thb.THcopy);
					THcopyCount ++;			
				}
			}while (thb != thElement)
			
			for (let i = 0; i < thElement.WSFfollows.length;i++){
						thStart.before(thElement.WSFfollows[i]);
						if (thElement.WSFfollows[i].THcopy != null){
							if (thStartCopy != null) thStartCopy.before(thElement.WSFfollows[i].THcopy);
							else firstRow.THcopy.append(thElement.WSFfollows[i].THcopy);
							THcopyCount ++;			
						}										
			}
		
			if (THcopyCount > 0){
				if (this.tabelle.WSTableHeader != null)
					this.tabelle.WSTableHeader.TFloesche(THcopyCount,lastRow.THcopy);
			}
			lastRow.className = lastRow.WSFoldClass; 
				

		}			
			
		let icon = this.newIconZu(thElement); 
		icon.addEventListener('click',this.hideRows);
		thElement.replaceChild(icon,thElement.firstChild);
		
		if (this.storeSession){
			window.sessionStorage.setItem(this.storeText(thElement),'open');
		}
/*		
		// Kopie aus TableHeader
		if (thElement.THCopy != null){
			let thc = thElement.THCopy;
			let icon = this.newIconZu(thElement); 
			icon.addEventListener('click',this.hideRows);
			thc.replaceChild(icon,thc.firstChild);
		}
// */		
        return;
	}

	/**
	Schalter zum Aufklappen erzeugen
	@param th TH-Element, in welches der Schalter später eingetragen wird.
	 */	
	newIconAuf(th) {
		var newElement = document.createElement('span');
		if (this.iconOpenInCSS )
			newElement.innerHTML = '<span title="aufklappen" class="WSFIcon WSFIconOpen"></span>';
		else 
			newElement.innerHTML = '<span title="aufklappen" class="WSFIcon WSFIconOpen">&dArr;</span>';
		newElement.WSFTableFolder = this;
		newElement.WSFth=th;
		return newElement;			
	}

	/**
	Schalter zum Einklappen erzeugen
	@param th TH-Element, in welches der Schalter später eingetragen wird.
	 */	
	newIconZu(th) {
		var newElement = document.createElement('span');
		if (this.iconCloseInCSS)
			newElement.innerHTML = '<span  title="einklappen" class="WSFIcon WSFIconClose"></span>';
		else	
			newElement.innerHTML = '<span  title="einklappen" class="WSFIcon WSFIconClose">&rArr;</span>';
		newElement.WSFTableFolder = this;
		newElement.WSFth=th;
		return newElement;			
	}

	/**
	Eventhandler
	 */
	showRows (event){
		var target = event.target;
		while (target.WSFth == null) target = target.parentNode;
		var thElement = target.WSFth;	
		var tf = target.WSFTableFolder;
		tf.doShowRows(thElement,false);	
		if (tf.tabelle.WSTableHeader != null)
			tf.tabelle.WSTableHeader.resize();
	}

	hideRows (event){
		var target = event.target;
		while (target.WSFth == null) target = target.parentNode;
		var thElement = target.WSFth;	
		var tf = target.WSFTableFolder;
		tf.doHideRows(thElement,false);		
		if (tf.tabelle.WSTableHeader != null)
			tf.tabelle.WSTableHeader.resize();
	}

	/**
	Für ganze Spalten
	 */
	hideColRows (event){
		var target = event.target;
		while (target.WSFth == null) target = target.parentNode;
		var thElement = target.WSFth;	
		var tf = target.WSFTableFolder;
		for (let i = tf.thElemente.length -1; i >= 0; i--){
			let p = thElement.WSFpos;	
			if (tf.thElemente[i].WSFpos >= p
			    && tf.thElemente[i].WSFopen){
					tf.doHideRows(tf.thElemente[i],false);		
			}
		}
		
		var rowElements = thElement.parentElement.children;
		var start = false;
		for (let p = 0; p < rowElements.length; p++){
			let element = rowElements.item(p);
			if (element == thElement) start = true;
			let text = 'TH '+element.innerText;
			if (start 
				&& (element.classList.contains('WSFSummen') 
					|| element.classList.contains('WSFHeader'))){
				let icon = tf.newIconAuf(element); 
				icon.addEventListener('click',tf.showColRows);
				element.replaceChild(icon,element.firstChild);			
			}			
		}
		if (tf.tabelle.WSTableHeader != null)
			tf.tabelle.WSTableHeader.resize();
	}

	showColRows (event){
		var target = event.target;
		while (target.WSFth == null) target = target.parentNode;
		var thElement = target.WSFth;	
		var tf = target.WSFTableFolder;

		for (let i = 0; i <  tf.thElemente.length; i++){
			let p = thElement.WSFpos;	
			if (tf.thElemente[i].WSFpos <= p
			    && ! tf.thElemente[i].WSFopen){
				tf.doShowRows(tf.thElemente[i],false);		
			}
		}

		var rowElements = thElement.parentElement.children;
		for (let p = 0; p < rowElements.length; p++){
			let element = rowElements.item(p);
			let text = 'TH '+element.innerText;
			if (element.classList.contains('WSFSummen') 
				|| element.classList.contains('WSFHeader')){
				let icon = tf.newIconZu(element); 
				icon.addEventListener('click',tf.hideColRows);
				element.replaceChild(icon,element.firstChild);			
			}			
			if (element == thElement) p = rowElements.length; 
		}
		if (tf.tabelle.WSTableHeader != null)
			tf.tabelle.WSTableHeader.resize();
	}
	
	storeText(thElement){
		var text = "WSTF:"+thElement.children.item(1).innerText;
		var th = thElement.WSTprev;
		while (th != null){
			text += ':'+th.children.item(1).innerText;
			th = th.WSTprev;
		} 
		return text;
	}
	
	restore(){
		if (window.sessionStorage != null){
			this.storeSession = false;
			for (var i = 0; i < this.thElemente.length; i++){
				var thText = this.storeText(this.thElemente[i]);
				if (thText != null){
					var status = window.sessionStorage.getItem(thText);
					if (status != null){
						if (status == 'open'  && ! this.thElemente[i].WSFopen) this.doShowRows(this.thElemente[i]);
						else
						if (status == 'closed'  && this.thElemente[i].WSFopen) this.doHideRows(this.thElemente[i]);
					} 								
				}
			}
			if (this.tabelle.WSTableHeader != null) this.tabelle.WSTableHeader.resize();
			this.storeSession = true;			
			console.log('Hinweis: Der TableFolder speichert den Zustand der Knoten')
		}else console.log('window.sessionStorage ist null');		
	}
}